const defaultAcceptedProbabilityThreshold=0.8;

const kMinLengthForToken=3;

const kMaxLengthForToken=12;

const M_LN2=0.69314718055994530942;

const M_E = 2.7182818284590452354;

var ds=	Components.classes['@mozilla.org/file/directory_service;1'].
		getService(Components.interfaces.nsIProperties);

var tokenRegex=new RegExp("[ \\t\\n\\r\\f.,]");

var observerService =Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);

function compareDistance(a,b){
	if(a.distance<b.distance) return -1;
	else if(a.distance>b.distance) return 1;
	return 0;
}

function isDigit(car){
	return car>='0' && car<='9';
}

function isDecimalNumber(word)
{
    var i=0;
    if(word[i]=="-") i++;
    while(i<word.length){
	if(!isDigit(word[i])) return false;
	i++;
    } 
    return true;
}

function getScriptableStream(is){
	var sis=Components.classes["@mozilla.org/scriptableinputstream;1"].createInstance(Components.interfaces.nsIScriptableInputStream);
	sis.init(is);
	return sis;
}

function isASCII(word)
{
    var i=0;
    while(i<word.length){
	if(word[i]>127) return false;
	i++;
    }
    return true;
}

function Token(word,count,probability){
	this.word=word;
	this.probability=probability;
	this.count=count;
	this.getWord=function(){return this.word;}
	this.getProbability=function(){return this.probability;}
	this.getCount=function(){return this.count;}
}

function TokenEnumeration(table){
	this.table=table;
	this.aSize=table.length;
	this.indice=0;
	this.hasMoreTokens=function(){
		return this.indice<this.aSize;
	}
	this.nextToken=function(){
		if(this.indice>=this.aSize) throw "Indice fuori dai limiti";
		var ret=this.table[this.indice];
		this.indice++;
		return ret;
	}
}

function Tokenizer(){
	this.tokenTable=new Object();
	this.tokenTable.length=0;
	this.clearTokens=function(){
		this.tokenTable=new Object();
		this.tokenTable.length=0;
	}
	this.add=function(word,count){
		if(!count) count=1;
		token=this.tokenTable[word];
		if(token){
			token.count+=count;
		}
		else{
			token=new Token(word,count,0);
			this.tokenTable[word]=token;
			this.tokenTable.length++;
		}
	}
	this.tokenizeHeaders=function(names,values){
		while(names.hasMore()){
			var name=names.getNext();
			name=name.toLowerCase();
			var value=values.getNext();
			switch(name[0]){
				case 'c':
					if(name=="content-type"){
					
					}
					break;
				case 'r':
					if(name=="received"){

					}
					break;
				case 's':
					if(name=="subject"){
						this.addTokenForHeader(name,value,true);
					}
					else if(name=="sender"){
						this.tokenizeAddress(value,"sender");
					}
					break;
				case 'x':
					break;
				case 'u':
					break;
				default:
					break;
			}
		}
	}
	this.addTokenForHeader=function(prefix,text,tokenize){
		if(!tokenize){
			this.add(prefix+":"+text.toLowerCase());
			return;
		}
		var word="";
		var i=0;
		while(i<text.length){
			if(text[i].search(tokenRegex)==-1){
				word+=text[i];	
			}
			else if(word!=""){
				if(isDecimalNumber(word)){}
				else if (isASCII(word)) this.add(prefix+":"+word.toLowerCase());				
				word="";
			}
			i++;
		}
		if(word!=""){
			if(isDecimalNumber(word)){}
			else if (isASCII(word)) this.add(prefix+":"+word.toLowerCase());			
		}		
	}
	this.tokenizeAttachment=function(contentType,filename){
		contentType=contentType.toLowerCase();
		filename=filename.toLowerCase();
		this.addTokenForHeader("attachment/filename", filename);
		this.addTokenForHeader("attachment/content-type",contentType);
	}
	this.remove=function(word,count){
		if(!count) count=1;
		token=this.tokenTable[word];
		if(token){
			if(token.count>count){
				token.count-=count;
			}
			else if(token.count==count) delete this.tokenTable[word];
		}
	}
	this.stripHTML=function(text){
		var textStripped=text.replace(/<[^>]+>/g,"");
		return textStripped;
	}
	this.countTokens=function(){
		return this.tokenTable.length;
	}
	this.tokenizeAddress=function(address,prefix){
		if(!prefix) prefix="";
		else prefix=prefix+":";
		var len=address.length;
		var ind=address.search('@');
		if (ind < len - 1){
			this.add(prefix+address.substring(0,ind));
			this.add(prefix+address.substring(ind+1));
		}
		else{
			this.add(prefix+address);
		}
	}
	this.tokenize_ascii_word=function(word){
		word=word.toLowerCase();
		var lun=word.length;
		if(lun >= kMinLengthForToken && lun <= kMaxLengthForToken){
			this.add(word);
		}
		else if(lun > kMaxLengthForToken){
			    if (lun < 40 && word.split('@').length-1 == 1){
      					var ind=word.search('@'); 
      					if (ind < lun - 1) // if the @ sign is the last character, it must not be an email address
					{
 						this.tokenizeAddress(word);
      					}
    				}
		}
	}
	this.getTokens=function(){
		var ret=new Array();
		var i=0;
		for(x in this.tokenTable){
			if(this.tokenTable[x] instanceof Token){
				ret[i]=this.tokenTable[x];
				i++;
			}
		}
		return new TokenEnumeration(ret);
	}
	this.get=function(word){
		if(word){	
			var token=this.tokenTable[word];
			if(!token) return null;
			return token; 
		}
		return null;
	}
	this.set=function(word,count){
		var token=this.tokenTable[word];
		if(token!=null){
			token.count=count;
		}
		else{
			token=new Token(word,count,0);
			this.tokenTable[word]=token;
		}	
	}
	this.canc=function(word){
		this.tokenTable[word]=null;
	}
	this.tokenize=function(text){
		text=this.stripHTML(text);
		var word="";
		var i=0;
		while(i<text.length){
			if(text[i].search(tokenRegex)==-1){
				word+=text[i];	
			}
			else if(word!=""){
				if(isDecimalNumber(word)){}
				else if (isASCII(word)) this.tokenize_ascii_word(word);					
				word="";
			}
			i++;
		}
		if(word!=""){
			if(isDecimalNumber(word)){}
			else if (isASCII(word)) this.tokenize_ascii_word(word);					
		}
	}
	this.QueryInterface= function (iid) {
        	if (!iid.equals(Components.interfaces.nsIMCTokenizer)
        	    && !iid.equals(Components.interfaces.nsISupports)){
        		    throw Components.results.NS_ERROR_NO_INTERFACE;
       		 }
      		 return this;
   	}
}


function nsMCFilter(){
	this.acceptedThreshold = 0;
	this.acceptedProbabilityThreshold = this.acceptedThreshold/ 100;
	if (this.acceptedProbabilityThreshold == 0 || this.acceptedProbabilityThreshold >= 1)
      		this.acceptedProbabilityThreshold = defaultAcceptedProbabilityThreshold;
	observerService.addObserver(this,"xpcom-shutdown",false);
	this.subFolders=new Object();	
}

nsMCFilter.prototype = {
	MAX_MESSAGES: 1000000,
	MODE_RDWR: 0x20 | 0x04 ,
	MODE_RDONLY: 0x01,
	setFolder: function(folder){
		var profileFolder = ds.get('ProfD', Components.interfaces.nsIFile);	
		var servername=folder.server.realHostName;
		var username = folder.server.realUsername;
		var file=profileFolder;
		file.append("MailClassifier");
		file.append(username+"@"+servername);
		var path=new Array();
		var currentFolder=folder;
		while(currentFolder!=folder.rootFolder && currentFolder){
			path.push(currentFolder.name);
			currentFolder=currentFolder.parentMsgFolder;
		} 
		for(var i=path.length-1;i>=0;i--) file.append(path[i]);
		file.append(folder.name+".mctf");
		this.file=file;
		if(!file.exists()){
			try{
				file.create(0,0666);
				this.writeTrainingData();
			}
			catch(e){throw "cannot_create"}
		}
		
	},
	addSubFolder: function(folderURI){
		this.subFolders[folderURI]=new Object();
		this.subFolders[folderURI].count=0;
		this.subFolders[folderURI].tokens=new Tokenizer();
	},
	init: function(){
		this.readTrainingData();
	},

	classifyMessage: function(uriM,aMsgWindow,listener,tokenizer){
			/***
			* l'algoritmo utilizza il teorema di bayes, per cui se indichiamo
			* con A l'evento "questa cartella" e con x e y due generici eventi "nell'email c'è il token...", 
			* ipotizzando che x e y sia stocasticamente indipendenti, allora
			*
			* 		                      P(x|A) * P(y|A) * P(A)
			*  P(A | x e y) =  __________________________________________________________
			*
			*		   P(A) * P(x|A) * P(y|A) + P(non A) * P(x|non A) * P(y|non A)
			* 
			***/
			/*if (listener && this.mGoodCount==0 && this.mGoodTokens.countTokens()==0) {
				listener.onMessageClassified(uri, Components.interfaces.nsIMCFilter.BAD,tokenizer);
			     	return;
    			}
    			if (listener && this.mBadCount==0 && this.mBadTokens.countTokens()==0) {
      				listener.onMessageClassified(uri, Components.interfaces.nsIMCFilter.GOOD,tokenizer);
      				return;
    			}*/
			var i;
			var goodclues=0;
			
			var probF=new Object();	
			var nMessages=new Object();
			var nFolders=0;	

			var totMessages=0;
			for(uri in this.subFolders){
				var cou=this.subFolders[uri].count;
				nMessages[uri]=cou;
				totMessages+=cou;	
				nFolders++;	
			}			
			
			// qui calcola le probabilita
			for(uri in this.subFolders){
				probF[uri] = this.subFolders[uri].count/totMessages;
			}	

			var ite=tokenizer.getTokens();
			var tokens=new Array();

			while(ite.hasMoreTokens()){
				var t=ite.nextToken();
				var word=t.getWord();

				var t1=new Object();
				t1.word=t.getWord();
				t1.count=t.getCount();
				
				var distance=0;
				var probXF=new Object();
			
				for(uri in this.subFolders){			
					var token=this.subFolders[uri].tokens.get(word);
					// il numero di messaggi della cartella in cui compare il token					
					var folCount=((token!=null) ? token.getCount() : 0);
					var folNum=nMessages[uri];
					var probXThis=folCount/folNum;
					var distanceThis=Math.abs(probXThis - 0.5);
					distance = (distanceThis>distance) ? distanceThis : distance;

					if(folNum>0 && folCount>0) 
						probXF[uri] = probXThis;
					else
						probXF[uri] = 0.01;
				}
				
				// la distanza della probabilità da 0.5 
				if(distance>=0.1){
					goodclues++;
					t1.distance=distance;
					t1.prob=probXF;
				}
				else{
					t1.distance = -1; //ignora l'aiuto
				}
				tokens.push(t1);
			}
			tokens.sort(compareDistance);
			var count=tokens.length;
			var last=count;
			
			// tiene in considerazione al massimo i 150 token più significativi
			var first=(goodclues > 150) ? count - 150 : 0;

			
			var valueF=probF;
			
			for (i = first; i < last; i++){
				if (tokens[i].distance != -1){
					for(uri in tokens[i].prob){
						valueF[uri] *= tokens[i].prob[uri];
					}        				
            			}
			}

			var totValueF=0;
			for(uri in this.subFolders){
				totValueF += valueF[uri];
			}

			for(uri in this.subFolders){
				prob = 	valueF[uri]/totValueF;
				prob=Math.round(prob*100);
				if(listener) listener.onMessageClassified(uriM,uri,prob);							
			}
			if(listener) listener.endClassification(uriM);							
	},
	rememberTokens: function(corpus,tokens){
		while (tokens.hasMoreTokens()) {
        		var token = tokens.nextToken();
        		corpus.add(token.getWord(),1);
    		}
		//dump(this.mGoodCount+"\n");
		//dump(this.mBadCount+"\n");
		//this.dumpTokenizer(this.mGoodTokens);
		//dump("\n");
		//this.dumpTokenizer(corpus);
	},
	dumpTokenizer: function(tok){
		var tokens=tok.getTokens();
		while(tokens.hasMoreTokens()){
			var token=tokens.nextToken();
			dump(token.word+","+token.count);
			if(tokens.hasMoreTokens()) dump(",");
		}
	},
	setMessageClassification: function(tokenizer,uri,old,nuo,listener){		
		var tokens = tokenizer.getTokens();		
		var fol=this.subFolders[nuo];
		if(fol){
			fol.count++;
			this.rememberTokens(fol.tokens,tokens);
		}		

		if(listener)
			listener.endTraining(uri);							
	},
	readTrainingData: function(){
		if(!this.file) return;
		try{
			var inp = Components.classes["@mozilla.org/network/file-input-stream;1"].createInstance(Components.interfaces.nsIFileInputStream);
			inp.init(this.file,this.MODE_RDONLY,0666,1);
			inp=getScriptableStream(inp);
			while(inp.available()){			
				var uri=this.readWord(inp)[0];
				var count=this.readWord(inp)[0];
				count=parseInt(count);
				var tokens=new Tokenizer();
				this.readTokens(inp,tokens);
				if(this.subFolders[uri] && !isNaN(count) && count>=0){
					this.subFolders[uri].count=count;
					this.subFolders[uri].tokens=tokens;
				}
			}	
			inp.close();
		}
		catch(e){
			dump(e);
			this.file.remove(false);
		}
	},
	normalizeTokenizer: function(tokenizer,actual,next){
		var tokens=tokenizer.getTokens();
		while(tokens.hasMoreTokens()){
			var token=tokens.nextToken();
			var count=token.count;
			count=Math.round(count/actual*next);
			if(count>0) tokenizer.set(token.word,count);
			else tokenizer.canc(token.word);
		}
	},
	normalizeData: function(){
		var mass=null;
		for(uri in this.subFolders){
			if(!mass && this.subFolders[uri].count>=this.MAX_MESSAGES){
				mass=uri;
			}
			if(mass && this.subFolders[uri].count>this.subFolders[mass].count) mass=uri;
		}
		if(mass!=null){
			var newMax=this.MAX_MESSAGES/10;
			var oldMax=this.subFolders[mass].count;
			this.subFolders[mass].count=newMax;	
			this.normalizeTokenizer(this.subFolders[mass].tokens,oldMax,newMax);					
			for(uri in this.subFolders){
				if(uri!=mass){
					var oldCount=this.subFolders[uri].count;
					var newCount=Math.floor(oldCount/oldMax*newMax);	
					this.subFolders[uri].count=newCount;
					this.normalizeTokenizer(this.subFolders[uri].tokens,oldCount,newCount);
				}
			}
		}
	},
	writeTrainingData:  function(){
		if(!this.file) return;
		try{
			this.normalizeData();
			var out = Components.classes["@mozilla.org/network/file-output-stream;1"].createInstance(Components.interfaces.nsIFileOutputStream);
			out.init(this.file,this.MODE_RDWR, 420, 0 );
			for(uri in this.subFolders){			
				var subF=this.subFolders[uri];
				var write=uri+"\n";
				out.write(write,write.length);			
				var write=subF.count+"\n";
				out.write(write,write.length);
				this.writeTokens(out,subF.tokens);
			}
			out.close();
		}
		catch(e){
			dump(e);
			this.file.remove(false);
		}
	},
	writeTokens: function(out,tokenizer){
		var ret;
		var tokenCount=tokenizer.countTokens();
		if(tokenCount>0){
			var tokens=tokenizer.getTokens();
			var tokensA=new Array();
			while(tokens.hasMoreTokens()){
				var token=tokens.nextToken();
				tokensA.push(token);
			}
			tokensA.sort(function(a,b){
				if(a.getCount()>b.getCount()) return -1;
				else if(a.getCount()<b.getCount()) return 1;
				else return 0;
			});
			var len=300;
			if(tokensA.length<len) len=tokensA.length;

			for(var i=0;i<len;i++){
				var token=tokensA[i];
				ret=out.write(token.word,token.word.length);
				if(ret!=token.word.length) throw "a";
				
				ret=out.write(",",1);
				if(ret!=1) throw "e";
				
				var write=token.count+"";
				ret=out.write(write,write.length);
				if(ret!=write.length) throw "b";

				if((i+1)<len){
					ret=out.write(",",1);
					if(ret!=1) throw "c";
				}			
			}
		}
		ret=out.write("\n",1);
		if(ret!=1) throw "d";
	},
	readWord: function(inp){
		var word="";
		var ch=null;
		while(inp.available() && ch!='\n' && ch!=','){
			ch=inp.read(1);
			if(ch!='\n' && ch!=','){
				word+=ch;
			}
		}
		if(ch!='\n' && ch!=',' && ch!=null) throw "a";
		return [word,ch];
	},
	readTokens: function(inp,tokenizer){
		var ch=",";
		while(ch!="\n"){
			var word=this.readWord(inp);
			if(word[0]=="") break;
			if(word[1]=="\n") throw "b";
			var count=this.readWord(inp);
			var countInt=parseInt(count[0]);
			if(!isNaN(countInt) && countInt>0 && word[0]){
				tokenizer.add(word[0],countInt);
			}
			ch=count[1];
		}
	},
   	classifyMessages: function(aCount,
                          aMsgURIs,
                          aMsgWindow,
                          aListener,
			   tokenizer){
		var i;
		for(i=0;i<aMsgURIs.length;i++){
			this.classifyMessage(aMsgURIs[i],aMsgWindow,aListener,tokenizer);
		}
	},
    	resetTrainingData: function(){
		for(uri in this.subFolders){
			this.removed(uri);
			this.addSubFolder(uri);			
		}
	},
	removed: function(folder){
		delete this.subFolders[folder];
	},
	observe: function(subject,topic,data){
		if(topic=="xpcom-shutdown"){
			this.shutdown();
		}
	},
	shutdown: function(){
		this.writeTrainingData();
	},

    	QueryInterface: function (iid) {
        	if (!iid.equals(Components.interfaces.nsIMCFilter)
        	    && !iid.equals(Components.interfaces.nsISupports)){
        		    throw Components.results.NS_ERROR_NO_INTERFACE;
       		 }
      		 return this;
   	}
}

var Module = {
    firstTime: true,

    registerSelf: function (compMgr, fileSpec, location, type) {
        if (this.firstTime) {
            this.firstTime = false;
            throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
        }
        compMgr =
	compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
        compMgr.registerFactoryLocation(this.filterCID,
                                        "nsMCFilter",
                                        this.filterID,
                                        fileSpec,
                                        location,
                                        type);
        compMgr.registerFactoryLocation(this.tokenizerCID,
                                        "nsMCTokenizer",
                                        this.tokenizerID,
                                        fileSpec,
                                        location,
                                        type);
    },

    getClassObject : function (compMgr, cid, iid) {
        if (!cid.equals(this.filterCID) && !cid.equals(this.tokenizerCID))
        throw Components.results.NS_ERROR_NO_INTERFACE
        if (!iid.equals(Components.interfaces.nsIFactory))
        throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
        if(cid.equals(this.filterCID)) return this.filterFactory;
    	if(cid.equals(this.tokenizerCID)) return this.tokenizerFactory;
    },

    filterCID: Components.ID("{01c379ec-16a2-4c3d-84f3-1779f9ea7485}"),
    filterID: "@uniroma3.it/MailClassifier/nsMCFilter;1",
    tokenizerCID: Components.ID("{621e6307-6d93-4f23-ac36-4d472c0b4de8}"),
    tokenizerID: "@uniroma3.it/MailClassifier/nsMCTokenizer;1",

    filterFactory: {
        createInstance: function (outer, iid) {
            if (outer != null)
            throw Components.results.NS_ERROR_NO_AGGREGATION;
		return (new nsMCFilter()).QueryInterface(iid);
        }
    },
    tokenizerFactory: {
        createInstance: function (outer, iid) {
            if (outer != null)
            throw Components.results.NS_ERROR_NO_AGGREGATION;
		return (new Tokenizer()).QueryInterface(iid);
        }
    },

    canUnload: function(compMgr) {
        return true;
    }
}; // END Module

function NSGetModule(compMgr, fileSpec) { return Module; }
